#!/usr/bin/perl
# $Id: podbrowser.pl,v 1.8 2004/09/06 14:44:09 jodrell Exp $
# Copyright (c) 2004 Gavin Brown. All rights reserved. This program is free
# software; you can redistribute it and/or modify it under the same terms
# as Perl itself.

use Gtk2 -init;
use Gtk2::GladeXML;
use Gtk2::SimpleList;
use Gtk2::PodViewer 0.08;
use Gtk2::PodViewer::Parser qw(decode_entities);
use Locale::gettext;
use POSIX qw(setlocale);
use strict;

### set up global variables:
my $NAME		= 'podbrowser';
my $VERSION		= '0.02';
my $PREFIX		= '@PREFIX@';
my $GLADE_FILE		= (-d $PREFIX ? sprintf('%s/share/%s', $PREFIX, lc($NAME)) : $ENV{PWD}).sprintf('/%s.glade', lc($NAME));
my $LOCALE_DIR		= (-d $PREFIX ? "PREFIX/share/locale" : $ENV{PWD}.'/locale');

my $SEARCH_OFFSET	= 0;
my $RCFILE		= sprintf('%s/.%src', $ENV{HOME}, lc($NAME));
my $MAXIMIZED		= 0;
my $FULLSCREEN		= 0;
my $OPTIONS		= load_config();
my @HISTORY		= split(/\|/, $OPTIONS->{history});
my @FORWARD;
my @BACK;
my $CURRENT_DOCUMENT;
my $LAST_SEARCH_STR;

### set up l10n support:
setlocale(LC_ALL, $ENV{LANG});
bindtextdomain(lc($NAME), $LOCALE_DIR);
textdomain(lc($NAME));

### bits we'll be reusing:
chomp(my $OPENER	= `which gnome-open 2> /dev/null`);
my $APP			= Gtk2::GladeXML->new($GLADE_FILE);
my $TIPS		= Gtk2::Tooltips->new;
my $THEME		= get_an_icon_theme();

my $IDX_PBF		= $APP->get_widget('main_window')->render_icon('gtk-jump-to', 'menu');
my $PAGE_PBF		= Gtk2::Gdk::Pixbuf->new_from_file($THEME->lookup_icon('gnome-mime-text', 16, 'force-svg')->get_filename)->scale_simple(16, 16, 'bilinear');
my $FOLDER_PBF		= Gtk2::Gdk::Pixbuf->new_from_file($THEME->lookup_icon('gnome-fs-directory', 16, 'force-svg')->get_filename)->scale_simple(16, 16, 'bilinear');
my $NORMAL_CURSOR	= Gtk2::Gdk::Cursor->new('left_ptr');
my $BUSY_CURSOR		= Gtk2::Gdk::Cursor->new('watch');
my $ITEMS		= generate_index();
my %categories		= (
	funcs		=> gettext('Functions'),
	modules		=> gettext('Modules'),
	pragma		=> gettext('Pragmas'),
	pods		=> gettext('POD Documents'),
);

### start building the UI:

$APP->signal_autoconnect_from_package('main');
$APP->get_widget('location')->disable_activate;
$APP->get_widget('open_dialog_location')->disable_activate;

my $viewer = Gtk2::PodViewer->new;
$viewer->signal_connect('link_clicked' => \&link_clicked);
$viewer->signal_connect('link_enter', sub { set_status($_[1]) });
$viewer->signal_connect('link_leave', sub { set_status('') });
$APP->get_widget('viewer_scrwin')->add($viewer);
$viewer->show;

### build a SimpleList from the glade widget for the document index:
my $page_index	= Gtk2::SimpleList->new_from_treeview(
	$APP->get_widget('document_index'),
	icon	=> 'pixbuf',
	mark	=> 'text',
	link	=> 'hidden',
);

$page_index->get_selection->signal_connect('changed', sub {
	my $idx = ($page_index->get_selected_indices)[0];
	my $mark = $page_index->{data}[$idx][2];
	$viewer->jump_to($mark);
	return 1;
});

### build a tree widget for the full index:
my $model = Gtk2::TreeStore->new(qw/Gtk2::Gdk::Pixbuf Glib::String/);
$APP->get_widget('index')->set_model($model);
$APP->get_widget('index')->insert_column_with_attributes(
	0,
	'pixbuf',
	Gtk2::CellRendererPixbuf->new,
	pixbuf => 0,
);
$APP->get_widget('index')->insert_column_with_attributes(
	1,
	'document',
	Gtk2::CellRendererText->new,
	text => 1,
);

### populate the tree:
foreach my $category (sort keys %{$ITEMS}) {
	if ($category ne '') {
		my $parent_iter = $model->append(undef);
		$model->set($parent_iter, 0, $FOLDER_PBF);
		$model->set($parent_iter, 1, $categories{$category});
		foreach my $doc (sort keys %{$ITEMS->{$category}}) {
			if ($doc ne '') {
				my $iter = $model->append($parent_iter);
				$model->set($iter, 0, $PAGE_PBF);
				$model->set($iter, 1, $doc);
			}
		}
	}
}

$APP->get_widget('index')->get_selection->signal_connect('changed', \&index_changed);

$APP->get_widget('main_window')->set_icon(Gtk2::Gdk::Pixbuf->new_from_file($THEME->lookup_icon(lc($NAME), 16, 'force-svg')->get_filename));

### if the program was run with an argument, load the argument as a document:
if ($ARGV[0] ne '') {
	Glib::Timeout->add(50, sub {
		set_location($ARGV[0]);
		return undef;
	});
}

Glib::Timeout->add(50, \&timeout);

### apply the user's previous preferences:
$APP->get_widget('pane')->set_position($OPTIONS->{pane_position}	|| 200);
$APP->get_widget('vpane')->set_position($OPTIONS->{vpane_position}	|| 250);
$APP->get_widget('show_index')->set_active($OPTIONS->{show_index} == 0 && defined($OPTIONS->{show_index}) ? undef : 1);
$APP->get_widget('location')->set_popdown_strings('', @HISTORY);
$APP->get_widget('main_window')->maximize if ($OPTIONS->{maximized} == 1);
$APP->get_widget('main_window')->set_default_size($OPTIONS->{window_x}, $OPTIONS->{window_y});

$APP->get_widget('main_window')->signal_connect('window-state-event', \&window_changed_state);

Gtk2->main;

exit;

### this runs every 50ms, and keeps track of the state of the navigation buttons and tooltips:
sub timeout {
	### the back button:
	if (scalar(@BACK) > 0 && !$APP->get_widget('back_button')->get('sensitive')) {
		$APP->get_widget('back_button')->set_sensitive(1);
		$TIPS->set_tip($APP->get_widget('back_button'), sprintf(gettext("Go back to '%s'"), $BACK[scalar(@BACK) - 1]));

	} elsif (scalar(@BACK) < 1 && $APP->get_widget('back_button')->get('sensitive')) {
		$APP->get_widget('back_button')->set_sensitive(undef);
		$TIPS->set_tip($APP->get_widget('back_button'), '');

	}

	### the forward button:
	if (scalar(@FORWARD) > 0 && !$APP->get_widget('forward_button')->get('sensitive')) {
		$APP->get_widget('forward_button')->set_sensitive(1);
		$TIPS->set_tip($APP->get_widget('forward_button'), sprintf(gettext("Go forward to '%s'"), $FORWARD[0]));

	} elsif (scalar(@FORWARD) < 1 && $APP->get_widget('forward_button')->get('sensitive')) {
		$APP->get_widget('forward_button')->set_sensitive(undef);
		$TIPS->set_tip($APP->get_widget('forward_button'), '');
	}

	### the up button:
	if ($APP->get_widget('location')->entry->get_text =~ /::/) {
		$APP->get_widget('up_button')->set_sensitive(1);
		my @parts = split(/::/, $APP->get_widget('location')->entry->get_text);
		pop(@parts);
		my $doc = join('::', @parts);
		$TIPS->set_tip($APP->get_widget('up_button'), sprintf(gettext("Go up to '%s'"), $doc));

	} else {
		$APP->get_widget('up_button')->set_sensitive(undef);
		$TIPS->set_tip($APP->get_widget('up_button'), '');

	}

	### the go button:
	if ($APP->get_widget('location')->entry->get_text ne '' && !$APP->get_widget('go_button')->get('sensitive')) {
		$APP->get_widget('go_button')->set_sensitive(1);

	} elsif ($APP->get_widget('location')->entry->get_text eq '' && $APP->get_widget('go_button')->get('sensitive')) {
		$APP->get_widget('go_button')->set_sensitive(undef);

	}

	return 1;
}

### pops up the 'open document' dialog:
sub open_dialog {
	$APP->get_widget('open_dialog')->set_icon($APP->get_widget('main_window')->get_icon);
	$APP->get_widget('open_dialog')->show_all;
	$APP->get_widget('open_dialog_location')->set_popdown_strings($APP->get_widget('location')->entry->get_text, @HISTORY);
	return 1;
}

### handles the 'open document' dialog response:
sub open_dialog_response {
	if ($_[1] eq 'ok' || $_[1] == 1) {
		set_location($APP->get_widget('open_dialog_location')->entry->get_text);
	}
	$APP->get_widget('open_dialog')->hide_all;
	return 1;
}
sub open_dialog_delete_event {
	$APP->get_widget('open_dialog')->hide_all;
	return 1;
}
sub on_open_dialog_location_activate {
	$APP->get_widget('open_dialog')->signal_emit('response', 1);
	return 1;
}
sub browse_button_clicked {
	my $dialog = Gtk2::FileChooserDialog->new(
		gettext('Choose File'),
		undef,
		'open',
		'gtk-cancel'	=> 'cancel',
		'gtk-ok'	=> 'ok'
	);
	$dialog->signal_connect('response', sub {
		if ($_[1] eq 'ok') {
			$APP->get_widget('open_dialog_location')->entry->set_text($dialog->get_filename);
		}
		$dialog->destroy;
	});
	$dialog->set_icon($APP->get_widget('main_window')->get_icon);
	$dialog->run;
	return 1;
}

### shows/hides the left pane of the window:
sub toggle_index {
	if ($_[0]->get_active) {
		$APP->get_widget('vpane')->show_all;
	} else {
		$APP->get_widget('vpane')->hide_all;
	}
	$OPTIONS->{show_index} = ($_[0]->get_active ? 1 : 0);

	return 1;
}

sub about {
	my $about = Gtk2::Dialog->new;
	$about->set_resizable(0);
	$about->set_icon($APP->get_widget('main_window')->get_icon);
	my $label = Gtk2::Label->new;
	my $markup = <<"END";
<span weight="bold" size="large">POD Browser</span>

Using Gtk2::PodViewer v$Gtk2::PodViewer::VERSION

Copyright (c) 2004
Gavin Brown
Torsten Schoenfeld
Scott Arrington

<span size="small">This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.</span>
END
	$label->set_markup($markup);
	$label->set_justify('center');
	$label->set_line_wrap(1);
	$about->set_title('About Pod Viewer');
	$about->set_border_width(8);
	$about->add_buttons('gtk-close' => 'close');
	$about->vbox->pack_start(Gtk2::Image->new_from_stock('gtk-dialog-info', 'dialog'), 0, 0, 0);
	$about->vbox->pack_start($label, 1, 1, 0);
	$about->show_all;
	$about->signal_connect('response', sub { $about->destroy; });
	$about->show_all;;
	return 1;
}

### this is used when the user requests a new document, via the location entry,
### the go button, the 'open document' dialog, the index or via a clicked link:
sub go {
	my $text = $APP->get_widget('location')->entry->get_text;

	$APP->get_widget('main_window')->window->set_cursor($BUSY_CURSOR);
	$viewer->get_window('text')->set_cursor($BUSY_CURSOR);
	Gtk2->main_iteration while (Gtk2->events_pending);

	if (!$viewer->load($text)) {
		$APP->get_widget('location')->entry->set_text($CURRENT_DOCUMENT);
		pop(@BACK);
		$APP->get_widget('main_window')->window->set_cursor($NORMAL_CURSOR);
		$viewer->get_window('text')->set_cursor($NORMAL_CURSOR);
		my $dialog = Gtk2::MessageDialog->new($APP->get_widget('main_window'), 'modal', 'error', 'ok', sprintf(gettext("Couldn't find a POD document for '%s'."), $text));
		$dialog->signal_connect('response', sub { $dialog->destroy });
		$dialog->run;

	} else {
		$CURRENT_DOCUMENT = $text;
		$APP->get_widget('main_window')->set_title($text);
		$APP->get_widget('main_window')->window->set_cursor($NORMAL_CURSOR);
		$viewer->get_window('text')->set_cursor($NORMAL_CURSOR);

		### populate the index:
		@{$page_index->{data}} = ();
		map { push(@{$page_index->{data}}, [ $IDX_PBF, section_reformat(decode_entities($_)), $_ ]) } $viewer->get_marks;
		unshift(@HISTORY, $text);

		### update the history, removing duplicates:
		my %seen;
		for (my $i = 0 ; $i < scalar(@HISTORY) ; $i++) {
			if ($seen{$HISTORY[$i]} == 1) {
				splice(@HISTORY, $i, 1);
			} else {
				$seen{$HISTORY[$i]} = 1;
			}
		}
		$APP->get_widget('location')->set_popdown_strings(@HISTORY);
		$LAST_SEARCH_STR = '';

		### this goes through the site index, looking to see if the requested document is present, and
		### selects it if found:
		my $found = 0;
		for (my $i = 0 ; $i < scalar(keys(%categories)) ; $i++) {
			my $category = (grep { $_ ne '' } sort(keys(%categories)))[$i];
			for (my $j = 0 ; $j < scalar(keys(%{$ITEMS->{$category}})) ; $j++) {
				my $doc = (grep { $_ ne '' } sort(keys(%{$ITEMS->{$category}})))[$j];
				if ($doc eq $text) {
					$found++;
					$APP->get_widget('index')->expand_row(Gtk2::TreePath->new_from_indices($i), undef);

					unless ($APP->get_widget('index')->has_focus) {
						$APP->get_widget('index')->get_selection->select_path(Gtk2::TreePath->new_from_indices($i, $j));
						$APP->get_widget('index')->scroll_to_cell(Gtk2::TreePath->new_from_indices($i, ($j <= 4 ? 0 : $j - 4)), undef, 1, 0, 0);
					}
				}
			}
		}
		if ($found == 0) {
			$APP->get_widget('index')->collapse_all;
			$APP->get_widget('index')->get_selection->unselect_all;
		}
	}

	return 1;
}

sub close_window {
	if ($MAXIMIZED == 1) {
		$OPTIONS->{maximized} = 1;
	} else {
		$OPTIONS->{maximized} = 0;
		($OPTIONS->{window_x}, $OPTIONS->{window_y}) = $APP->get_widget('main_window')->get_size if ($FULLSCREEN == 0);
	}

	$OPTIONS->{pane_position} = $APP->get_widget('pane')->get_position;
	$OPTIONS->{vpane_position} = $APP->get_widget('vpane')->get_position;
	$OPTIONS->{history} = join('|', @HISTORY);
	save_config();
	exit 0;
}

sub load_config {
	my $OPTIONS = {};
	if (open(RCFILE, $RCFILE)) {
		while (<RCFILE>) {
			chomp;
			my ($name, $value) = split(/\s*=\s*/, $_, 2);
			$OPTIONS->{lc($name)} = $value;
		}
		close(RCFILE);
	}
	return $OPTIONS;
}

sub save_config {
	if (!open(RCFILE, ">$RCFILE")) {
		printf(STDERR "Cannot open file '%s' for writing: %s\n", $RCFILE, $!);
		return undef;
	} else {
		foreach my $key (sort(keys(%{$OPTIONS}))) {
			printf(RCFILE "%s=%s\n", $key, $OPTIONS->{$key});
		}
		close(RCFILE);
		return 1;
	}
	return undef;
}

sub link_clicked {
	my (undef, $text) = @_;
	$text =~ s/\"$//g;
	$text =~ s/^\"//g;

	return undef if ($text eq '');

	my @marks = $viewer->get_marks;
	my $seen = 0;
	map { s/^[\"\']//g ; s/[\"\']$//g ; $seen++ if (lc($_) eq lc($text)) } @marks;

	if ($seen > 0) {
		# link referred to an anchor:
		for (my $i = 0 ; $i < scalar(@marks) ; $i++) {
			$marks[$i] =~ s/^[\"\']//g;
			$marks[$i] =~ s/[\"\']$//g;
			if (lc($marks[$i]) eq lc($text)) {
				$page_index->select($i);
				return 1;
			}
		}

	} elsif ($text =~ /\|\/?/) {
		# link referred to an anchor, but with some named text:
		my ($text, $section) = split(/\|\/?/, $text, 2);
		link_clicked(undef, $section);

	} elsif ($text =~ /^(\w+)\:\/\//) {
		# link referred to a URL:
		open_url($text);

	} elsif ($text =~ /^\// && ! -e $text) {
		# link referred to a non-existent file, remove the leading slash and try again:
		$text =~ s/^\///;
		link_clicked(undef, $text);

	} elsif ($text =~ /\// && ! -e $text) {
		# link referred to a poddoc/anchor anchor, split the text and try with the second part:
		my ($doc, $section) = split(/\//, $text, 2);
		set_location($doc);
		link_clicked(undef, $section);

	} else {
		# link referred to another pod document:
		set_location($text);
	}
	return 1;

}

sub set_status {
	my $str = shift;
	$APP->get_widget('status_bar')->push($APP->get_widget('status_bar')->get_context_id($str), $str);
	return 1;
}

sub set_location {
	my $locn = shift;
	if ($APP->get_widget('location')->entry->get_text ne '') {
		push(@BACK, $APP->get_widget('location')->entry->get_text);
	}
	@FORWARD = ();
	$APP->get_widget('location')->entry->set_text($locn);
	go();
}

sub go_back {
	unshift(@FORWARD, $APP->get_widget('location')->entry->get_text);
	my $locn = pop(@BACK);
	$APP->get_widget('location')->entry->set_text($locn);
	go();
}

sub go_forward {
	push(@BACK, $APP->get_widget('location')->entry->get_text);
	my $locn = shift(@FORWARD);
	$APP->get_widget('location')->entry->set_text($locn);
	go();
}

sub go_up {
	my @parts = split(/::/, $APP->get_widget('location')->entry->get_text);
	pop(@parts);
	$APP->get_widget('location')->entry->set_text(join('::', @parts));
	push(@BACK, $CURRENT_DOCUMENT);
	@FORWARD = ();
	go();
}

sub user_set_location {
	return undef if ($APP->get_widget('location')->entry->get_text eq '');
	push(@BACK, $CURRENT_DOCUMENT) if ($CURRENT_DOCUMENT ne '' && $CURRENT_DOCUMENT ne $APP->get_widget('location')->entry->get_text);
	@FORWARD = ();
	go();
}

sub search {
	my $str = $APP->get_widget('search_entry')->get_text;

	$str =~ s/^\s*$//g;

	return undef if ($str eq '');

	set_status('Searching...');

	$APP->get_widget('main_window')->window->set_cursor($BUSY_CURSOR);
	$viewer->get_window('text')->set_cursor($BUSY_CURSOR);

	my $doc = $viewer->get_buffer->get_text(
		$viewer->get_buffer->get_start_iter,
		$viewer->get_buffer->get_end_iter,
		1
	);

	$str = quotemeta($str);
	$APP->get_widget('search_entry')->set_sensitive(0);

	$SEARCH_OFFSET = 0 if ($str ne $LAST_SEARCH_STR);
	$LAST_SEARCH_STR = $str;

	for ($SEARCH_OFFSET ; $SEARCH_OFFSET < length($doc) ; $SEARCH_OFFSET++) {
		Gtk2->main_iteration while (Gtk2->events_pending);
		if (substr($doc, $SEARCH_OFFSET) =~ /^$str/i) {
			my $iter = $viewer->get_buffer->get_iter_at_offset($SEARCH_OFFSET);
			$viewer->scroll_to_iter($iter, undef, 1, 0, 0);
			$APP->get_widget('search_entry')->set_sensitive(1);
			$APP->get_widget('search_entry')->grab_focus();
			$viewer->get_buffer->move_mark(
				$viewer->get_buffer->get_mark('insert'), 
				$viewer->get_buffer->get_iter_at_offset($SEARCH_OFFSET)
			);
			$viewer->get_buffer->move_mark(
				$viewer->get_buffer->get_mark('selection_bound'), 
				$viewer->get_buffer->get_iter_at_offset($SEARCH_OFFSET + length($str))
			);

			set_status('');
			$APP->get_widget('main_window')->window->set_cursor($NORMAL_CURSOR);
			$SEARCH_OFFSET += length($str);
			return 1;
		}
	}
	$APP->get_widget('search_entry')->set_sensitive(1);
	$APP->get_widget('search_entry')->grab_focus();

	set_status('');

	$SEARCH_OFFSET = 0;

	$APP->get_widget('main_window')->window->set_cursor($NORMAL_CURSOR);
	$viewer->get_window('text')->set_cursor($NORMAL_CURSOR);

	my $dialog = Gtk2::MessageDialog->new($APP->get_widget('main_window'), 'modal', 'info', 'ok', sprintf(gettext("The string '%s' was not found."), $str));
	$dialog->signal_connect('response', sub { $dialog->destroy });
	$dialog->show_all;

	return undef;
}

sub select_all {
	$viewer->get_buffer->move_mark(
		$viewer->get_buffer->get_mark('insert'), 
		$viewer->get_buffer->get_start_iter,
	);
	$viewer->get_buffer->move_mark(
		$viewer->get_buffer->get_mark('selection_bound'), 
		$viewer->get_buffer->get_end_iter,
	);
	return 1;
}

sub search_dialog {
	$APP->get_widget('search_dialog_entry')->set_text($APP->get_widget('search_entry')->get_text);
	$APP->get_widget('search_dialog')->show_all;
	return 1;
}

sub search_dialog_close {
	$APP->get_widget('search_dialog')->hide_all;
	return 1;
}

sub search_dialog_entry_activate() {
	search_dialog_response(undef, 'ok');
	return 1;
}

sub search_dialog_response {
	if ($_[1] eq 'ok') {
		$APP->get_widget('search_entry')->set_text($APP->get_widget('search_dialog_entry')->get_text);
		search();

	}

	$APP->get_widget('search_dialog')->hide_all;
	return 1;
}

sub open_url {
	my $url = shift;

	if (!-x $OPENER) {
		my $dialog = Gtk2::MessageDialog->new($APP->get_widget('main_window'), 'modal', 'info', 'ok', gettext("Cannot find the gnome-open program to launch this URL was not found."));
		$dialog->signal_connect('response', sub { $dialog->destroy });
		$dialog->show_all;
		return undef;

	} else {
		system("$OPENER \"$url\" &");
		return 1;

	}
}

### this looks through the system for perl pod documents that reference functions, modules and pod
### documents, and creates a hash of hashes that can be used to build a sitewide index:
sub generate_index {
	my %PATHS;

	# doing a reverse sort means that later versions of Perl are preferred over newer versions:
	foreach my $dir (reverse sort @INC) {
		if (-r "$dir/pod/perltoc.pod" && $PATHS{perltoc} eq '') {
			$PATHS{perltoc} = "$dir/pod/perltoc.pod";
		}

		if (-r "$dir/pod/perlfunc.pod" && $PATHS{perlfunc} eq '') {
			$PATHS{perlfunc} = "$dir/pod/perlfunc.pod";
		}

		if (-r "$dir/perllocal.pod" && $PATHS{perllocal} eq '') {
			$PATHS{perllocal} = "$dir/perllocal.pod";
		}
	}

	my $ITEMS = {};
	my $category;

	if (-r $PATHS{perltoc}) {
		if (!open(PERLTOC, $PATHS{perltoc})) {
			print STDERR "$PATHS{perltoc}: $!\n";

		} else {
			while (<PERLTOC>) {
				if (/BASIC DOCUMENTATION/) {
					$category = 'pods';

				} elsif (/PRAGMA DOCUMENTATION/) {
					$category = 'pragma';

				} elsif (/MODULE DOCUMENTATION/) {
					$category = 'modules';

				} elsif (/AUXILIARY DOCUMENTATION/) {
					$category = 'aux';

				} elsif (/^=head2/) {
					my (undef, $doc, undef) = split(' ', $_, 3);
					$doc =~ s/[^A-Za-z0-9\:\_]//g;
					$ITEMS->{$category}->{$doc}++;
				}
			}
			close(PERLTOC);
		}
	}

	$category = '';
	if (-r $PATHS{perlfunc}) {
		if (!open(PERLFUNC, $PATHS{perlfunc})) {
			print STDERR "$PATHS{perlfunc}: $!\n";

		} else {
			while (<PERLFUNC>) {
				if (/Alphabetical Listing of Perl Functions/) {
					$category = 'funcs';
				} elsif (/^=item/) {
					my (undef, $doc, undef) = split(' ', $_, 3);
					$doc =~ s/[^A-Za-z0-9\_\/\-]+//g;
					$ITEMS->{$category}->{$doc}++;
				}
			}
			close(PERLFUNC);
		}
	}

	if (-r $PATHS{perllocal}) {
		if (!open(PERLLOCAL, $PATHS{perllocal})) {
			print STDERR "$PATHS{perllocal}: $!\n";

		} else {
			while (<PERLLOCAL>) {
				if (/^=head2 .+? L<([A-Za-z0-9\:]+)\|/) {
					my $module = $1;
					$module =~ s/[^A-Za-z0-9\:\_]//g;
					$ITEMS->{modules}->{$1}++;
				}
			}
			close(PERLLOCAL);
		}
	}

	return $ITEMS;
}

sub section_reformat {
	my $str = shift;
	my @words = split(/[\s\t]+/, $str);
	my @return = '';
	foreach my $word (@words) {
		if ($word =~ /^[A-Z]+$/) {
			$word = ucfirst(lc($word));
		}
		push(@return, $word);
	}
	return join(' ', @return);
}

### this tracks the window's state:
sub window_changed_state {
	my $mask = $_[1]->changed_mask;
	if ("$mask" eq '[ withdrawn ]') {
		$MAXIMIZED  = 0;
		$FULLSCREEN = 0;
	} elsif ("$mask" eq '[ maximized ]') {
		if ($MAXIMIZED == 1) {
			$MAXIMIZED  = 0;
		} else {
			$MAXIMIZED  = 1;
		}
		$FULLSCREEN = 0;
	} elsif ("$mask" eq '[ fullscreen ]') {
		$MAXIMIZED  = 0;
		if ($FULLSCREEN == 1) {
			$FULLSCREEN  = 0;
		} else {
			$FULLSCREEN  = 1;
		}
	}
	return 1;
}

### this is called when the user clicks on an item in the site index:
sub index_changed {
	my ($path) = $APP->get_widget('index')->get_selection->get_selected_rows;
	if (defined($path)) {
		my $path = $path->to_string;
		my $iter = $APP->get_widget('index')->get_model->get_iter_from_string($path);
		my $value = $APP->get_widget('index')->get_model->get_value($iter, 1);
		my $seen = 0;
		foreach my $category (keys %categories) {
			$seen++ if ($categories{$category} eq $value);
		}
		if ($seen < 1) {
			$APP->get_widget('location')->entry->set_text($value);
			go();
		}
	}
	return 1;
}

### returns an icon theme. When running in a GNOME session, the default is fine. if not,
### we have to load a custom one:
sub get_an_icon_theme {
	my $theme = Gtk2::IconTheme->get_default;
	if ($theme->get_example_icon_name eq '') {
		$theme = Gtk2::IconTheme->new;
		$theme->set_custom_theme($OPTIONS->{theme} ne '' ? $OPTIONS->{theme} : 'gnome');
	}
	return $theme;
}
